from __future__ import annotations

from typing import TYPE_CHECKING

# noinspection PyProtectedMember
from officialeye._api.mutator import IMutator

# noinspection PyProtectedMember
from officialeye._api.template.matcher import IMatcher
from officialeye._api_builtins.interpretation.file import FileInterpretation
from officialeye._api_builtins.interpretation.file_temp import FileTempInterpretation
from officialeye._api_builtins.interpretation.ocr_tesseract import TesseractInterpretation
from officialeye._api_builtins.matcher.sift_flann import SiftFlannMatcher
from officialeye._api_builtins.mutator.clahe import CLAHEMutator
from officialeye._api_builtins.mutator.grayscale import GrayscaleMutator
from officialeye._api_builtins.mutator.non_local_means_denoising import NonLocalMeansDenoisingMutator
from officialeye._api_builtins.mutator.rotate import RotateMutator
from officialeye._api_builtins.supervisor.combinatorial import CombinatorialSupervisor
from officialeye._api_builtins.supervisor.least_squares_regression import LeastSquaresRegressionSupervisor

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    from officialeye._api.context import Context

    # noinspection PyProtectedMember
    from officialeye._api.template.interpretation import IInterpretation

    # noinspection PyProtectedMember
    from officialeye._api.template.supervisor import ISupervisor
    from officialeye.types import ConfigDict


"""
Mutator generators
"""


def _gen_mutator_grayscale(config: ConfigDict, /) -> IMutator:
    return GrayscaleMutator(config)


def _gen_mutator_non_local_means_denoising(config: ConfigDict, /) -> IMutator:
    return NonLocalMeansDenoisingMutator(config)


def _gen_mutator_clahe(config: ConfigDict, /) -> IMutator:
    return CLAHEMutator(config)


def _gen_mutator_rotate(config: ConfigDict, /) -> IMutator:
    return RotateMutator(config)


"""
Matcher generators
"""


def _gen_matcher_sift_flann(config: ConfigDict, /) -> IMatcher:
    return SiftFlannMatcher(config)


"""
Supervisor generators
"""


def _gen_supervisor_combinatorial(config: ConfigDict, /) -> ISupervisor:
    return CombinatorialSupervisor(config)


def _gen_supervisor_least_squares_regression(config: ConfigDict, /) -> ISupervisor:
    return LeastSquaresRegressionSupervisor(config)


"""
Interpretation generators
"""


def _gen_interpretation_file(config: ConfigDict, /) -> IInterpretation:
    return FileInterpretation(config)


def _gen_interpretation_file_temp(config: ConfigDict, /) -> IInterpretation:
    return FileTempInterpretation(config)


def _gen_interpretation_ocr_tesseract(config: ConfigDict, /) -> IInterpretation:
    return TesseractInterpretation(config)


def initialize_builtins(context: Context, /):

    # register mutators
    context.register_mutator(GrayscaleMutator.MUTATOR_ID, _gen_mutator_grayscale)
    context.register_mutator(NonLocalMeansDenoisingMutator.MUTATOR_ID, _gen_mutator_non_local_means_denoising)
    context.register_mutator(CLAHEMutator.MUTATOR_ID, _gen_mutator_clahe)
    context.register_mutator(RotateMutator.MUTATOR_ID, _gen_mutator_rotate)

    # register matchers
    context.register_matcher(SiftFlannMatcher.MATCHER_ID, _gen_matcher_sift_flann)

    # register supervisors
    context.register_supervisor(CombinatorialSupervisor.SUPERVISOR_ID, _gen_supervisor_combinatorial)
    context.register_supervisor(LeastSquaresRegressionSupervisor.SUPERVISOR_ID, _gen_supervisor_combinatorial)

    # register interpretations
    context.register_interpretation(FileInterpretation.INTERPRETATION_ID, _gen_interpretation_file)
    context.register_interpretation(FileTempInterpretation.INTERPRETATION_ID, _gen_interpretation_file_temp)
    context.register_interpretation(TesseractInterpretation.INTERPRETATION_ID, _gen_interpretation_ocr_tesseract)



from abc import ABC

from officialeye.error.codes import (
    ERR_TEMPLATE_ID_NOT_UNIQUE,
    ERR_TEMPLATE_INVALID_CONCURRENCY_CONFIG,
    ERR_TEMPLATE_INVALID_FEATURE,
    ERR_TEMPLATE_INVALID_FEATURE_CLASS,
    ERR_TEMPLATE_INVALID_INTERPRETATION,
    ERR_TEMPLATE_INVALID_KEYPOINT,
    ERR_TEMPLATE_INVALID_MATCHING_ENGINE,
    ERR_TEMPLATE_INVALID_MUTATOR,
    ERR_TEMPLATE_INVALID_SUPERVISION_ENGINE,
    ERR_TEMPLATE_INVALID_SYNTAX,
)
from officialeye.error.error import OEError
from officialeye.error.modules import ERR_MODULE_TEMPLATE


class ErrTemplate(OEError, ABC):

    def __init__(self, code: int, code_text: str, while_text: str, problem_text: str, /, *, is_regular: bool = False, **kwargs):
        super().__init__(ERR_MODULE_TEMPLATE, code, code_text, while_text, problem_text, is_regular=is_regular)


class ErrTemplateInvalidSupervisionEngine(ErrTemplate):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_TEMPLATE_INVALID_SUPERVISION_ENGINE[0], ERR_TEMPLATE_INVALID_SUPERVISION_ENGINE[1], while_text, problem_text, **kwargs)

        self._init_args = while_text, problem_text, *kwargs

    def __reduce__(self):
        return self.__class__, self._init_args


class ErrTemplateInvalidMatchingEngine(ErrTemplate):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_TEMPLATE_INVALID_MATCHING_ENGINE[0], ERR_TEMPLATE_INVALID_MATCHING_ENGINE[1], while_text, problem_text, **kwargs)

        self._init_args = while_text, problem_text, *kwargs

    def __reduce__(self):
        return self.__class__, self._init_args


class ErrTemplateIdNotUnique(ErrTemplate):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_TEMPLATE_ID_NOT_UNIQUE[0], ERR_TEMPLATE_ID_NOT_UNIQUE[1], while_text, problem_text, **kwargs)

        self._init_args = while_text, problem_text, *kwargs

    def __reduce__(self):
        return self.__class__, self._init_args


class ErrTemplateInvalidKeypoint(ErrTemplate):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_TEMPLATE_INVALID_KEYPOINT[0], ERR_TEMPLATE_INVALID_KEYPOINT[1], while_text, problem_text, **kwargs)

        self._init_args = while_text, problem_text, *kwargs

    def __reduce__(self):
        return self.__class__, self._init_args


class ErrTemplateInvalidFeature(ErrTemplate):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_TEMPLATE_INVALID_FEATURE[0], ERR_TEMPLATE_INVALID_FEATURE[1], while_text, problem_text, **kwargs)

        self._init_args = while_text, problem_text, *kwargs

    def __reduce__(self):
        return self.__class__, self._init_args


class ErrTemplateInvalidConcurrencyConfig(ErrTemplate):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_TEMPLATE_INVALID_CONCURRENCY_CONFIG[0], ERR_TEMPLATE_INVALID_CONCURRENCY_CONFIG[1], while_text, problem_text, **kwargs)

        self._init_args = while_text, problem_text, *kwargs

    def __reduce__(self):
        return self.__class__, self._init_args


class ErrTemplateInvalidSyntax(ErrTemplate):
    def __init__(self, while_text: str, problem_text: str, yml_error: str | None = None, /, **kwargs):
        super().__init__(
            ERR_TEMPLATE_INVALID_SYNTAX[0], ERR_TEMPLATE_INVALID_SYNTAX[1], while_text, problem_text, **kwargs)

        self._init_args = while_text, problem_text, yml_error, *kwargs

        self.yml_error = yml_error

    def get_details(self) -> str | None:
        return self.yml_error

    def __reduce__(self):
        return self.__class__, self._init_args


class ErrTemplateInvalidFeatureClass(ErrTemplate):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_TEMPLATE_INVALID_FEATURE_CLASS[0], ERR_TEMPLATE_INVALID_FEATURE_CLASS[1], while_text, problem_text, **kwargs)

        self._init_args = while_text, problem_text, *kwargs

    def __reduce__(self):
        return self.__class__, self._init_args


class ErrTemplateInvalidMutator(ErrTemplate):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_TEMPLATE_INVALID_MUTATOR[0], ERR_TEMPLATE_INVALID_MUTATOR[1], while_text, problem_text, **kwargs)

        self._init_args = while_text, problem_text, *kwargs

    def __reduce__(self):
        return self.__class__, self._init_args


class ErrTemplateInvalidInterpretation(ErrTemplate):
    def __init__(self, while_text: str, problem_text: str, /, **kwargs):
        super().__init__(
            ERR_TEMPLATE_INVALID_INTERPRETATION[0], ERR_TEMPLATE_INVALID_INTERPRETATION[1], while_text, problem_text, **kwargs)

        self._init_args = while_text, problem_text, *kwargs

    def __reduce__(self):
        return self.__class__, self._init_args



# needed to not break type annotations if we are not in type checking mode
from __future__ import annotations

from types import TracebackType
from typing import TYPE_CHECKING, Dict

from officialeye._internal.feedback.abstract import AbstractFeedbackInterface
from officialeye._internal.feedback.dummy import DummyFeedbackInterface
from officialeye._internal.feedback.verbosity import Verbosity
from officialeye.error.error import OEError
from officialeye.error.errors.general import ErrInvalidKey
from officialeye.error.errors.template import ErrTemplateIdNotUnique

if TYPE_CHECKING:
    # noinspection PyProtectedMember
    # noinspection PyProtectedMember
    from officialeye._api.mutator import IMutator
    from officialeye._api.template.interpretation import IInterpretation

    # noinspection PyProtectedMember
    from officialeye._api.template.matcher import IMatcher

    # noinspection PyProtectedMember
    from officialeye._api.template.supervisor import ISupervisor
    from officialeye._internal.template.internal_template import InternalTemplate
    from officialeye.types import ConfigDict, InterpretationFactory, MatcherFactory, MutatorFactory, SupervisorFactory


class InternalContext:

    def __init__(self):
        self._afi = DummyFeedbackInterface()

        self._mutator_factories: Dict[str, MutatorFactory] = {}
        self._matcher_factories: Dict[str, MatcherFactory] = {}
        self._supervisor_factories: Dict[str, SupervisorFactory] = {}
        self._interpretation_factories: Dict[str, InterpretationFactory] = {}

        # keys: template ids
        # values: template
        self._loaded_templates: Dict[str, InternalTemplate] = {}

        # keys: paths to templates
        # values: corresponding template ids
        self._template_ids: Dict[str, str] = {}

    def setup(self, /, *, afi: AbstractFeedbackInterface, mutator_factories: Dict[str, MutatorFactory],
              matcher_factories: Dict[str, MatcherFactory], supervisor_factories: Dict[str, SupervisorFactory],
              interpretation_factories: Dict[str, InterpretationFactory]) -> InternalContext:
        assert afi is not None

        assert mutator_factories is not None
        assert matcher_factories is not None
        assert supervisor_factories is not None

        self._afi = afi
        self._mutator_factories = mutator_factories
        self._matcher_factories = matcher_factories
        self._supervisor_factories = supervisor_factories
        self._interpretation_factories = interpretation_factories

        return self

    def __enter__(self):
        return None

    def __exit__(self, exception_type: any, exception_value: BaseException | None, traceback: TracebackType | None):
        # inform the parent process that the current task is done
        self._afi.dispose(exception_type, exception_value, traceback)
        self._afi = DummyFeedbackInterface()

    def get_afi(self) -> AbstractFeedbackInterface:
        return self._afi

    def get_mutator(self, mutator_id: str, mutator_config: ConfigDict, /) -> IMutator:

        # TODO: (low priority) consider caching mutators that have the same id and configuration
        self._afi.info(Verbosity.DEBUG_VERBOSE, f"Loading mutator '{mutator_id}' with configuration {mutator_config}.")

        if mutator_id not in self._mutator_factories:
            raise ErrInvalidKey(
                f"while loading mutator '{mutator_id}'.",
                "Unknown mutator. Has this mutator been properly loaded?"
            )

        return self._mutator_factories[mutator_id](mutator_config)

    def get_matcher(self, matcher_id: str, matcher_config: ConfigDict, /) -> IMatcher:

        # TODO: (low priority) consider caching matchers that have the same id and configuration
        self._afi.info(Verbosity.DEBUG_VERBOSE, f"Loading matcher '{matcher_id}' with configuration {matcher_config}.")

        if matcher_id not in self._matcher_factories:
            raise ErrInvalidKey(
                f"while loading matcher '{matcher_id}'.",
                "Unknown matcher. Has this matcher been properly loaded?"
            )

        return self._matcher_factories[matcher_id](matcher_config)

    def get_supervisor(self, supervisor_id: str, supervisor_config: ConfigDict, /) -> ISupervisor:

        # TODO: (low priority) consider caching supervisors that have the same id and configuration
        self._afi.info(Verbosity.DEBUG_VERBOSE, f"Loading supervisor '{supervisor_id}' with configuration {supervisor_config}.")

        if supervisor_id not in self._supervisor_factories:
            raise ErrInvalidKey(
                f"while loading supervisor '{supervisor_id}'.",
                "Unknown supervisor. Has this supervisor been properly loaded?"
            )

        return self._supervisor_factories[supervisor_id](supervisor_config)

    def get_interpretation(self, interpretation_id: str, interpretation_config: ConfigDict, /) -> IInterpretation:

        # TODO: (low priority) consider caching interpretations that have the same id and configuration
        self._afi.info(Verbosity.DEBUG_VERBOSE, f"Loading interpretation '{interpretation_id}' with configuration {interpretation_config}.")

        if interpretation_id not in self._interpretation_factories:
            raise ErrInvalidKey(
                f"while loading interpretation '{interpretation_id}'.",
                "Unknown interpretation. Has this interpretation method been properly loaded?"
            )

        return self._interpretation_factories[interpretation_id](interpretation_config)

    def add_template(self, template: InternalTemplate, /):

        template_path = template.get_path()

        assert template_path not in self._template_ids, "A template from the same path has already been loaded"

        if template.identifier in self._loaded_templates:
            raise ErrTemplateIdNotUnique(
                f"while loading template '{template.identifier}'",
                "A template with the same id has already been loaded."
            )

        self._loaded_templates[template.identifier] = template
        self._template_ids[template_path] = template.identifier

        try:
            template.validate()
        except OEError as err:
            # rollback the loaded template
            del self._loaded_templates[template.identifier]
            del self._template_ids[template_path]

            # reraise the cause
            raise err

    def get_template(self, template_id: str, /) -> InternalTemplate:
        assert template_id in self._loaded_templates, "Unknown template id"
        return self._loaded_templates[template_id]

    def get_template_by_path(self, template_path: str, /) -> InternalTemplate | None:

        if template_path not in self._template_ids:
            return None

        template_id = self._template_ids[template_path]

        return self.get_template(template_id)



from __future__ import annotations

from typing import TYPE_CHECKING, List

from rich.json import JSON
from rich.table import Table

# noinspection PyProtectedMember
from officialeye._api.detection import detect

# noinspection PyProtectedMember
from officialeye._api.image import Image

# noinspection PyProtectedMember
from officialeye._api.template.template import Template
from officialeye._cli.context import CLIContext

# noinspection PyProtectedMember
from officialeye._internal.feedback.verbosity import Verbosity

if TYPE_CHECKING:
    from officialeye.types import FeatureInterpretation


def do_run(context: CLIContext, /, *, target_path: str, template_paths: List[str], interpret_path: str | None, visualize: bool):
    # print OfficialEye logo and other introductory information (if necessary)
    context.print_intro()

    # TODO: implement visualization generation
    # TODO: update the example in the documentation

    api_context = context.get_api_context()

    target_image = Image(api_context, path=target_path)

    interpretation_target_image = target_image if interpret_path is None else Image(api_context, path=interpret_path)

    templates = [Template(api_context, path=template_path) for template_path in template_paths]

    result = detect(api_context, *templates, target=target_image)

    interpretation_result = result.interpret(target=interpretation_target_image)

    table = Table()

    table.add_column("Feature", justify="right")
    table.add_column("Interpretation", justify="left")

    for feature in interpretation_result.template.features:
        interpretation: FeatureInterpretation = interpretation_result.get_feature_interpretation(feature)

        interpretation_visualization = JSON.from_data(interpretation, indent=4)

        table.add_row(feature.identifier, interpretation_visualization)

    context.get_terminal_ui().echo(Verbosity.INFO, table)
